"""Git VCS plugin
"""

from pathlib import Path

from cppython_core.plugin_schema.vcs import VersionControl
from dulwich.errors import NotGitRepository
from dulwich.repo import Repo


class Git(VersionControl):
    """Git implementation hooks"""

    @staticmethod
    def name() -> str:
        """The VCS name

        Returns:
            The name
        """
        return "git"

    def is_repository(self, path: Path) -> bool:
        """Queries repository status of a path

        Args:
            path: The input path to query

        Returns:
            Whether the given path is a repository root
        """

        try:
            Repo(str(path))
            return True

        except NotGitRepository:
            return False

    def extract_version(self, path: Path) -> str:
        """Extracts the system's version metadata

        Args:
            path: The repository path

        Returns:
            A version
        """
        return "0.1.0"



"""Test the functions related to the internal interface implementation and the 'Interface' interface itself
"""

import pytest
from click.testing import CliRunner
from pytest_cppython.plugin import InterfaceUnitTests

from cppython.console.interface import Configuration, ConsoleInterface, cli
from tests.data.fixtures import CPPythonProjectFixtures


class TestCLIInterface(CPPythonProjectFixtures, InterfaceUnitTests[ConsoleInterface]):
    """The tests for our CLI interface"""

    @pytest.fixture(name="plugin_type", scope="session")
    def fixture_plugin_type(self) -> type[ConsoleInterface]:
        """A required testing hook that allows type generation

        Returns:
            An overridden interface type
        """
        return ConsoleInterface

    def test_config(self) -> None:
        """Verify that the configuration object can be constructed"""

        Configuration()

    def test_verbosity(self) -> None:
        """Test that verbosity is passed through to the CLI"""

        config = Configuration()

        runner = CliRunner()
        result = runner.invoke(cli, "-v info", obj=config, catch_exceptions=False)

        assert result.exit_code == 0

        assert config.configuration.verbosity



"""Project test fixtures for all cppython tests
"""

from typing import Any

import pytest
from cppython_core.schema import (
    CPPythonLocalConfiguration,
    PEP621Configuration,
    PyProject,
    ToolData,
)
from pytest_cppython.fixtures import CPPythonFixtures
from pytest_cppython.mock import MockGenerator, MockProvider


class CPPythonProjectFixtures(CPPythonFixtures):
    """Additional fixtures to help test projects"""

    @pytest.fixture(name="tool", scope="session")
    def fixture_tool(self, cppython_local_configuration: CPPythonLocalConfiguration) -> ToolData:
        """The tool data
        Args:
            cppython_local_configuration: The parameterized cppython table
        Returns:
            Wrapped CPPython data
        """
        return ToolData(cppython=cppython_local_configuration)

    @pytest.fixture(name="project")
    def fixture_project(self, tool: ToolData, pep621_configuration: PEP621Configuration) -> PyProject:
        """Parameterized construction of PyProject data
        Args:
            tool: The tool table with internal cppython data
            pep621_configuration: The project table
        Returns:
            All the data as one object
        """
        return PyProject(project=pep621_configuration, tool=tool)

    @pytest.fixture(name="project_with_mocks")
    def fixture_project_with_mocks(self, project: PyProject) -> dict[str, Any]:
        """Extension of the 'project' fixture with mock data attached
        Args:
            project: The input project
        Returns:
            All the data as a dictionary
        """
        mocked_pyproject = project.dict(by_alias=True)
        mocked_pyproject["tool"]["cppython"]["provider"][MockProvider.name()] = {}
        mocked_pyproject["tool"]["cppython"]["generator"][MockGenerator.name()] = {}
        return mocked_pyproject



"""Everything needed to build a CPPython project
"""

from importlib import metadata
from logging import Logger
from pathlib import Path
from typing import Any

from cppython_core.exceptions import ConfigError
from cppython_core.plugin_schema.generator import Generator, GeneratorT
from cppython_core.plugin_schema.provider import Provider, ProviderT
from cppython_core.plugin_schema.vcs import VersionControl
from cppython_core.resolution import (
    resolve_cppython,
    resolve_cppython_plugin,
    resolve_generator,
    resolve_pep621,
    resolve_project_configuration,
    resolve_provider,
)
from cppython_core.schema import (
    CoreData,
    CorePluginData,
    CPPythonGlobalConfiguration,
    CPPythonLocalConfiguration,
    PEP621Configuration,
    Plugin,
    ProjectConfiguration,
)


class PluginBuilder:
    """Collection of utilities to collect and build plugins"""

    def __init__(self, group: str, logger: Logger) -> None:
        self._group = group
        self._logger = logger

    def gather_entries(self) -> list[metadata.EntryPoint]:
        """Gather all the available entry points for the grouping

        Returns:
            List of entries
        """
        return list(metadata.entry_points(group=f"cppython.{self._group}"))

    def load(self, entry_points: list[metadata.EntryPoint]) -> list[type[Plugin]]:
        """Loads a set of entry points

        Args:
            entry_points: The entry points to load

        Raises:
            TypeError: If an entry point is not a subclass of the 'Plugin' type

        Returns:
            List of plugin types
        """

        plugins = []

        for entry_point in entry_points:
            plugin = entry_point.load()

            if not issubclass(plugin, Plugin):
                raise TypeError("The CPPython plugin must be an instance of Plugin")

            plugins.append(plugin)

        return plugins


class Builder:
    """Helper class for building CPPython projects"""

    def __init__(self, logger: Logger) -> None:
        self.logger = logger

    def generate_core_data(
        self,
        configuration: ProjectConfiguration,
        pep621_configuration: PEP621Configuration,
        cppython_configuration: CPPythonLocalConfiguration,
    ) -> CoreData:
        """_summary_

        Args:
            self: _description_
            configuration: TODO
            pep621_configuration: TODO
            cppython_configuration: TODO

        Raises:
            ConfigError: Raised if data cannot be parsed

        Returns:
            _description_
        """

        global_configuration = CPPythonGlobalConfiguration()

        project_data = resolve_project_configuration(configuration)

        try:
            pep621_data = resolve_pep621(pep621_configuration, configuration)

        except ConfigError:
            configuration.version = self.extract_vcs_version(configuration.pyproject_file.parent)
            pep621_data = resolve_pep621(pep621_configuration, configuration)

        cppython_data = resolve_cppython(cppython_configuration, global_configuration, project_data)

        return CoreData(project_data=project_data, pep621_data=pep621_data, cppython_data=cppython_data)

    def extract_vcs_version(self, path: Path) -> str:
        """_summary_

        Args:
            path: _description_

        Raises:
            TypeError: _description_
            TypeError: _description_

        Returns:
            _description_
        """

        if not (vcs_types := self.discover_vcs()):
            raise TypeError("No VCS plugin found")

        plugin = None
        for vcs_type in vcs_types:
            vcs = vcs_type()
            if vcs.is_repository(path):
                plugin = vcs
                break

        if not plugin:
            raise TypeError("No applicable VCS plugin found for the given path")

        return plugin.extract_version(path)

    def discover_providers(self) -> list[type[Provider]]:
        """Discovers plugin types
        Raises:
            TypeError: Raised if the Plugin type is not subclass of 'Provider'

        Returns:
            List of Provider types
        """
        provider_builder = PluginBuilder(Provider.group(), self.logger)

        # Gather provider entry points without any filtering
        provider_entry_points = provider_builder.gather_entries()
        provider_types = provider_builder.load(provider_entry_points)

        plugins = []

        for provider_type in provider_types:
            if not issubclass(provider_type, Provider):
                raise TypeError("The CPPython plugin must be an instance of Plugin")

            plugins.append(provider_type)

        return plugins

    def discover_generators(self) -> list[type[Generator]]:
        """Discovers plugin types
        Raises:
            TypeError: Raised if the Plugin type is not subclass of 'Generator'

        Returns:
            List of Generator types
        """
        generator_builder = PluginBuilder(Generator.group(), self.logger)

        # Gather generator entry points without any filtering
        generator_entry_points = generator_builder.gather_entries()
        generator_types = generator_builder.load(generator_entry_points)

        plugins = []

        for generator_type in generator_types:
            if not issubclass(generator_type, Generator):
                raise TypeError("The CPPython plugin must be an instance of Plugin")

            plugins.append(generator_type)

        return plugins

    def discover_vcs(self) -> list[type[VersionControl]]:
        """Discovers plugin types
        Raises:
            TypeError: Raised if the Plugin type is not subclass of 'VersionControl'

        Returns:
            List of VersionControl types
        """
        vcs_builder = PluginBuilder(VersionControl.group(), self.logger)

        # Gather vcs entry points without any filtering
        vcs_entry_points = vcs_builder.gather_entries()
        vcs_types = vcs_builder.load(vcs_entry_points)

        plugins = []

        for vcs_type in vcs_types:
            if not issubclass(vcs_type, VersionControl):
                raise TypeError("The CPPython plugin must be an instance of Plugin")

            plugins.append(vcs_type)

        return plugins

    def create_generator(
        self, plugin_types: list[type[GeneratorT]], core_data: CoreData, generator_configuration: dict[str, Any]
    ) -> GeneratorT:
        """_summary_

        Args:
            plugin_types: _description_
            core_data: _description_
            generator_configuration: TODO

        Raises:
            ConfigError: TODO

        Returns:
            _description_
        """

        plugin_type, table = None, None
        # Extract the first type that's readable
        for plugin_type in plugin_types:
            name = plugin_type.name()

            try:
                table = generator_configuration[name]
                break
            except KeyError:
                continue

        if plugin_type is None:
            raise ConfigError("plugin_types was empty")

        if table is None:
            raise ConfigError("No generator name found within the 'tool.cppython.generator' table")

        generator_data = resolve_generator(core_data.project_data)

        cppython_plugin_data = resolve_cppython_plugin(core_data.cppython_data, plugin_type)

        core_plugin_data = CorePluginData(
            project_data=core_data.project_data,
            pep621_data=core_data.pep621_data,
            cppython_data=cppython_plugin_data,
        )

        return plugin_type(generator_data, core_plugin_data, table)

    def create_providers(
        self, plugin_types: list[type[ProviderT]], core_data: CoreData, provider_configuration: dict[str, Any]
    ) -> list[ProviderT]:
        """Creates Providers from input data

        Args:
            plugin_types: TODO
            core_data: TODO
            provider_configuration: TODO

        Returns:
            TODO
        """

        plugins = []

        for plugin_type in plugin_types:
            name = plugin_type.name()
            table = provider_configuration[name]

            provider_data = resolve_provider(core_data.project_data)

            cppython_plugin_data = resolve_cppython_plugin(core_data.cppython_data, plugin_type)

            core_plugin_data = CorePluginData(
                project_data=core_data.project_data,
                pep621_data=core_data.pep621_data,
                cppython_data=cppython_plugin_data,
            )

            plugins.append(plugin_type(provider_data, core_plugin_data, table))

        return plugins



"""Unit tests for the cppython VCS plugin
"""

import pytest
from pytest_cppython.plugin import VersionControlUnitTests

from cppython.plugins.git import Git


class TestGitInterface(VersionControlUnitTests[Git]):
    """Unit tests for the Git VCS plugin"""

    @pytest.fixture(name="plugin_type", scope="session")
    def fixture_plugin_type(self) -> type[Git]:
        """A required testing hook that allows type generation

        Returns:
            The VCS type
        """
        return Git



"""A click CLI for CPPython interfacing
"""

from logging import getLogger
from pathlib import Path

import click
import tomlkit
from cppython_core.plugin_schema.interface import Interface
from cppython_core.schema import ProjectConfiguration

from cppython.project import Project


def _find_pyproject_file() -> Path:
    """Searches upward for a pyproject.toml file

    Returns:
        The found directory
    """

    # Search for a path upward
    path = Path.cwd()

    while not path.glob("pyproject.toml"):
        if path.is_absolute():
            assert (
                False
            ), "This is not a valid project. No pyproject.toml found in the current directory or any of its parents."

    path = Path(path)

    return path


class Configuration:
    """Click configuration object"""

    def __init__(self) -> None:
        self.interface = ConsoleInterface()

        self.logger = getLogger("cppython.console")

        path = _find_pyproject_file()
        file_path = path / "pyproject.toml"

        self.configuration = ProjectConfiguration(pyproject_file=file_path, version=None)

    def query_vcs(self) -> str:
        """Queries the VCS system for its version

        Returns:
            The version
        """

        return "TODO"

    def generate_project(self) -> Project:
        """_summary_

        Returns:
            _description_
        """

        pyproject_data = tomlkit.loads(self.configuration.pyproject_file.read_text(encoding="utf-8"))

        return Project(self.configuration, self.interface, pyproject_data)


# Attach our config object to click's hook
pass_config = click.make_pass_decorator(Configuration, ensure=True)


@click.group()
@click.option("-v", "--verbose", count=True, help="Print additional output")
@pass_config
def cli(config: Configuration, verbose: int) -> None:
    """entry_point group for the CLI commands

    Args:
        config: The CLI configuration object
        verbose: The verbosity level
    """
    config.configuration.verbosity = verbose


@cli.command()
@pass_config
def info(config: Configuration) -> None:
    """Prints project information

    Args:
        config: The CLI configuration object
    """

    version = config.query_vcs()
    config.logger.info("The VCS project version is: %s", version)


@cli.command()
@pass_config
def install(config: Configuration) -> None:
    """Install API call

    Args:
        config: The CLI configuration object
    """
    project = config.generate_project()
    project.install()


@cli.command()
@pass_config
def update(config: Configuration) -> None:
    """Update API call

    Args:
        config: The CLI configuration object
    """
    project = config.generate_project()
    project.update()


class ConsoleInterface(Interface):
    """Interface implementation to pass to the project"""

    @staticmethod
    def name() -> str:
        """Returns the name of the interface

        Returns:
            The name
        """
        return "console"

    def write_pyproject(self) -> None:
        """Write output"""



"""Test the integrations related to the internal interface implementation and the 'Interface' interface itself
"""

import pytest
from pytest_cppython.plugin import InterfaceIntegrationTests

from cppython.console.interface import ConsoleInterface


class TestCLIInterface(InterfaceIntegrationTests[ConsoleInterface]):
    """The tests for our CLI interface"""

    @pytest.fixture(name="plugin_type", scope="session")
    def fixture_plugin_type(self) -> type[ConsoleInterface]:
        """A required testing hook that allows type generation

        Returns:
            An overridden interface type
        """
        return ConsoleInterface
